import requests
import json
import os
import pathlib
from .card import *
from urllib.parse import unquote,urlparse


import urllib3
urllib3.disable_warnings()

class Pipefy:
    url          = "https://api.pipefy.com/graphql"
    graph_folder = os.path.join(pathlib.Path(__file__).parent.resolve(),"graphql")
    
    def __init__(self,org_id,token):
        self.org_id = org_id
        self.token = token
        self.headers = {"Authorization":f"Bearer {token}",'Content-Type': 'application/json'}

    def runQuery(self,query:str)->dict:
        query = {"query": query}
        req = requests.post(self.url,headers=self.headers,data=json.dumps(query))
        #VERIFICA SE O REQUEST DEU ERRO
        if req.status_code != 200:
            raise Exception(req.text)
        data = req.json()

        if "errors" in data:
            raise Exception(data["errors"][0]["message"])
        return req.json()

    def presignedUrl(self,file_name:str)->str:
        query = open(os.path.join(self.graph_folder,"presignedUrl.gql"),'r').read()
        query = query.replace("$org_id$",self.org_id)
        query = query.replace("$file_name$",file_name)

        data = self.runQuery(query)
        
        return data["data"]["createPresignedUrl"]["url"]
    
    def createAttachment(self,file_path:str)->str:
        #pega nome do arquivo
        f_name = os.path.basename(file_path)
        
        #cria uma variavel assinada
        url = self.presignedUrl(f_name)

        #------------------- ENVIA ARQUIVO PARA A URL
        
        req = requests.put(url, verify=False,data=open(file_path, 'rb').read(),headers={'Content-Type': 'application/octet-stream'})
        if req.status_code != 200:
            raise Exception(req.text)


        return unquote(url.split("?")[0].replace("https://pipefy-prd-us-east-1.s3.amazonaws.com/",""))

    def listStartFormFields(self,pipe_id):

        query = open(os.path.join(self.graph_folder,"listStartFormFields.gql"),'r').read()
        query = query.replace("$pipe_id$",pipe_id)

        data = self.runQuery(query)

        return data.get("data").get("pipe").get("start_form_fields")
    
    def formatCardFields(self,card_dict:dict):
        data = {x['field']['id']:x['value'] for x in card_dict["fields"]}
        card_dict["fields"] = data
        return card_dict

    def getCard(self,card_id):

        query = open(os.path.join(self.graph_folder,"getCard.gql"),'r').read()
        query = query.replace("$card_id$",card_id)

        data = self.runQuery(query)
        card = data.get("data").get("card")

        return Card(self,card)

    def listPipeLabels(self,pipe_id):

        query = open(os.path.join(self.graph_folder,"listPipeLabels.gql"),'r').read()
        query = query.replace("$pipe_id$",pipe_id)

        data = self.runQuery(query)

        return data.get("data").get("pipe").get("labels")
    
    def listPhaseFormFields(self,phase_id):

        query = open(os.path.join(self.graph_folder,"listPhaseFormFields.gql"),'r').read()
        query = query.replace("$phase_id$",phase_id)

        data = self.runQuery(query)

        return data.get("data").get("phase").get("fields")

    def updateCardLabels(self,card_id,label_ids:list[int|str]):

        query = open(os.path.join(self.graph_folder,"updateCardLabels.gql"),'r').read()
        query = query.replace("$card_id$",card_id)
        query = query.replace("$label_ids$",json.dumps(label_ids))

        data = self.runQuery(query)

        return "OK"
    
    def listCardsFromPhase(self,phase_id,nextPage=None,format_fields=True):
        
        nextPage = f'"{nextPage}"' if nextPage else 'null'
        query = open(os.path.join(self.graph_folder,"listCardsFromPhase.gql"),'r').read()
        query = query.replace("$phase_id$",phase_id)
        query = query.replace("$after$",nextPage)

        data = self.runQuery(query)
    
        cards = data["data"]["phase"]["cards"]
        next_page = cards["pageInfo"]["hasNextPage"]
        cards_filtered = [x.get("node") for x in cards["edges"]]
        cards_filtered = [Card(self,card) for card in cards_filtered]
        if next_page:
            return cards_filtered+self.listCardsFromPhase(phase_id=phase_id,nextPage=cards["pageInfo"]["endCursor"])
        
        return cards_filtered

    def listMembers(self):
        query = open(os.path.join(self.graph_folder,"listMembers.gql"),'r').read()
        query = query.replace("$org_id$",self.org_id)

        data = self.runQuery(query)["data"]["organizations"][0]["members"]

        return [x.get("user") for x in data]
        
    def createCard(self,card:NewCard):
        query = open(os.path.join(self.graph_folder,"createCard.gql"),'r').read()
        query = query.replace("$title$",card.__title__)
        query = query.replace("$pipe_id$",card.__pipeid__)

        fields = []
        for f in card.used_fields:
            if f[1].type in [str,int,float]:
                value = object.__getattribute__(card,f[0])
                fields.append(f'{{field_id:{json.dumps(f[0])},field_value:{json.dumps(value)} }}')
            if f[1].type == list:
                value = object.__getattribute__(card,f[0])
                if f[1].is_file_path:
                    values = []
                    for fl in value:
                        values.append(self.createAttachment(fl))
                    fields.append(f'{{field_id:{json.dumps(f[0])},field_value:{json.dumps(values)} }}')
                elif f[1].list_sub_type in [str,int,float]:
                    fields.append(f'{{field_id:{json.dumps(f[0])},field_value:{json.dumps(value)} }}')
        
        fields = "\n,".join(fields)
        query = query.replace("$fields$",fields)
        query = query.replace('"null"',"null")

        return self.runQuery(query)["data"]["createCard"]["card"]

    def deleteCard(self,card_id:str):
        query = f'mutation {{ N0 :deleteCard(input:{{id: "{card_id}"}}){{ clientMutationId }}}}'
        return self.runQuery(query)["data"]["N0"]["clientMutationId"]

    def moveCard(self,card_id:str,phase_id:str):
        query = f'mutation {{moveCardToPhase(input: {{card_id: "{card_id}", destination_phase_id: "{phase_id}"}}){{card{{id}}}}}}'
        return self.runQuery(query)


    def updateFieldValue(self,card_id:str,field_id:str,value):
        val = f'"{value}"' if isinstance(value,str) else value
        val = json.dumps(val) if isinstance(val,list) else val
        query = f'mutation {{updateFieldsValues(input: {{nodeId: "{card_id}", values: [{{fieldId: "{field_id}", value: {val}}}]}}){{success}}}}'
        return self.runQuery(query)
    
    def downloadFile(self, file_path,destination):
        req = requests.get(file_path,headers=self.headers,stream=True,verify=False)

        #VERIFICA SE O REQUEST DEU ERRO
        if req.status_code != 200:
            raise Exception(req.text)
        
        file_name = urlparse(req.url).path.split("/")[-1]
        file_addr = os.path.join(destination, file_name)

        with open(file_addr, "wb") as file:
            for chunk in req.iter_content(1024):
                file.write(chunk)

        return file_addr




import pathlib
import os
from dateutil.parser import isoparse

graph_folder = os.path.join(pathlib.Path(__file__).parent.resolve(),"graphql")

#========================= CREATE CARD =======================
class CardField:
    def __init__(self,type,is_file_path=False,list_sub_type=str,required=True,default=None) -> None:
        self.type         = type
        self.is_file_path = is_file_path
        self.list_sub_type= list_sub_type
        self.required     = required
        self.default      = default

    def validate_default(self):
        if self.default is None: return

        if self.type == str:
            if not isinstance(self.default,str):
                raise Exception("Invalid default value")
        if self.type == int:
            if not isinstance(self.default,int):
                raise Exception("Invalid default value")
        if self.type == float:
            if not (isinstance(self.default,float) or isinstance(self.default,int)) :
                raise Exception("Invalid default value")
        if self.type == list:
            if not isinstance(self.default,list):
                raise Exception("Invalid default value")
            for item in self.default:
                if item.__class__ not in [str,int,float]:
                    raise Exception(f"invalid default value  '{item}'")
                elif self.is_file_path and not os.path.isfile(item):
                    raise Exception(f"default file not found '{item}'")
                elif self.list_sub_type == str and not isinstance(item,str):
                    raise Exception(f"invalid default value '{item}'")
                elif self.list_sub_type == int and not isinstance(item,int):
                    raise Exception(f"invalid default value '{item}'")
                elif self.list_sub_type == float and not isinstance(item,float):
                    raise Exception(f"invalid default value '{item}'")


class NewCard:
    graph_folder = os.path.join(pathlib.Path(__file__).parent.resolve(),"graphql")
    used_fields = None

    def __new__(cls,*args,**kwargs):
        instance = super(NewCard, cls).__new__(cls)
        #instance.__init__(**kwargs)
        # if isinstance(instance, Filho):
        #     print("Filho foi instanciado")
        instance.validateRequired(**kwargs)
        instance.validateFields(**kwargs)
        instance.setFields(**kwargs)
        return instance
        # pass

    # def __init__(self,**kwargs):
    #     self.validateRequired(**kwargs)
    #     self.validateFields(**kwargs)
    #     self.setFields(**kwargs)
    #     pass

    def setFields(self,**kwargs):
        
        for key, value in kwargs.items():
            setattr(self,key,value)
        
        #SET DEFAULT FIELDS
        my_fields = [(x,object.__getattribute__(self,x)) for x in self.__dir__() if isinstance(object.__getattribute__(self,x),CardField)]
        default_fields = [x for x in my_fields if x[1].default]

        not_defined = [x for x in default_fields if x[0] not in kwargs.keys()]

        for item in not_defined:
            setattr(self,item[0],item[1].default)

    def validateRequired(self,**kwargs):
        if not "__title__" in self.__dir__(): raise Exception("please define __title__")
        if not "__pipeid__" in self.__dir__(): raise Exception("please define __pipeid__")

        my_fields = [(x,object.__getattribute__(self,x)) for x in self.__dir__() if isinstance(object.__getattribute__(self,x),CardField)]
        required_ones = [x[0] for x in my_fields if x[1].required and x[1].default is None]
        missing = [x for x in required_ones if not x in kwargs]
        self.used_fields = [x for x in my_fields if x[1].required or x[1].default is not None]
        if missing: raise Exception(f"required field '{missing[0]}' not found!")


    def validateFields(self,**kwargs):
        my_fields = [x for x in self.__dir__()]
        for key, value in kwargs.items():
            if not key in my_fields:
                raise Exception(f"field '{key}' not found!")
            if not isinstance(object.__getattribute__(self,key),CardField):
                raise Exception(f"invalid field '{key}'")
            else:
                field = object.__getattribute__(self,key)
                if value.__class__ not in [str,int,list,float,int]:
                    raise Exception(f"invalid field type '{field.type}'")
                if field.type == str and not isinstance(value,str):
                    raise Exception(f"invalid value '{key}'")
                if field.type == float and not (isinstance(value,float) or isinstance(value,int)):
                    raise Exception(f"invalid value '{key}'")
                if field.type == int and not isinstance(value,int):
                    raise Exception(f"invalid value '{key}'")
                elif field.is_file_path and isinstance(value,str) and not os.path.isfile(value):
                    raise Exception(f"file not found '{value}'")
                elif field.type == list:
                    if not isinstance(value,list):
                        raise Exception(f"invalid field '{key}'")
                    for item in value:
                        if item.__class__ not in [str,int,list,float]:
                            raise Exception(f"invalid field type '{field.type}'")
                        elif field.is_file_path and not os.path.isfile(item):
                            raise Exception(f"file not found '{item}'")
                        elif field.list_sub_type == str and not isinstance(item,str):
                            raise Exception(f"invalid value '{item}'")
        pass





class Card:
    __pfy__ = None

    created_at = None
    created_by = None
    labels     = None
    due_date   = None
    id         = None
    fields     = None
    comments   = None

    __raw__ = None

    def __init__(self,pfy,dados:dict):
        self.__pfy__    = pfy
        self.__raw__    = dados.copy()
        self.id         = dados.pop('id')
        self.created_at = isoparse(dados.pop('created_at'))
        self.created_by = dados.pop('createdBy')
        self.due_date   = isoparse(dados.pop('due_date')) if dados['due_date'] else dados.pop('due_date')
        self.fields     = {x['field']['id']:x['value'] for x in dados.pop('fields')}
        self.comments   = [Comment(self.__pfy__,x) for x in dados.pop('comments')]
        self.labels     = dados.pop('labels')

        pass

    def newComment(self,text):
        query = open(os.path.join(graph_folder,"newComment.gql"),'r').read()
        query = query.replace("$card_id$",self.id)
        query = query.replace("$text$",text)

        data = self.__pfy__.runQuery(query)
        self.comments = [Comment(self.__pfy__,data["data"]["createComment"]["comment"])] + self.comments
        return self.comments[0]


        

    
    def __repr__(self):
        return f'Card<{self.id}>'


class Comment:
    __pfy__ = None
    
    id          = None
    author_id   = None
    author_name = None
    created_at  = None
    text        = None

    __raw__ = None

    def __init__(self,pfy,dados):
        self.__pfy__    = pfy
        self.__raw__    = dados
        self.id          = dados['id']
        self.author_id   = dados['author']['id']
        self.author_name = dados['author_name']
        self.created_at  = isoparse(dados['created_at'])
        self.text        = dados['text']
        pass

    def delete(self):
        query = open(os.path.join(graph_folder,"deleteComment.gql"),'r').read()
        query = query.replace("$comment_id$",self.id)
        self.__pfy__.runQuery(query)
        

    def __repr__(self):
        return f'Comment<{self.id}>'



#TEST FILE
from dotenv import load_dotenv
import sys
sys.path.append("./pipefyUts")
from pipefyUts import Pipefy,NewCard,CardField
import os
import json

load_dotenv(override=True)
ORG_ID =  os.getenv("ORG_ID")
TOKEN  = os.getenv("TOKEN")

pfy = Pipefy(ORG_ID,TOKEN)

a = pfy.getCard(card_id="1118627367")
#a.newComment("teste comentario")


#fl = pfy.listCardsFromPhase("325917747")[0]["fields"][3]["value"]
#fl = json.loads(fl)[0]

#pfy.downloadFile(file_path=fl,destination=".")

a = pfy.listCardsFromPhase("337516328")

flds = pfy.listPhaseFormFields("325917746")




class DefaultCardInfo(NewCard):
    #DEFAULT
    __title__                = "reembolso_compradores"
    __pipeid__               = "1025104"

    #PIPEFY FIELDS
    tipo_de_solicita_o              = CardField(str,default="Solicitação")
    solicitante1                    = CardField(list,default=["401961749"])
    categoria                       = CardField(str,default="Comprador/Pedido")
    explique_o_que_voc_precisa      = CardField(str,default="Solicitação de Reembolso aberta para pedidos > 180 dias.")
    cnpj                            = CardField(str,default="null")
    n_do_evento_obrigat_rio         = CardField(str)
    tipo_de_conta                   = CardField(str)
    nome_da_pessoa_titular_da_conta = CardField(str)
    banco                           = CardField(str)
    c_digo_do_banco_n_o_obrigat_rio = CardField(str)
    ag_ncia_sem_digito_obrigat_rio  = CardField(str)
    conta_sem_digito_obrigat_rio    = CardField(str)
    digito_da_conta_obrigat_rio     = CardField(str)


class PerdaFinanceiraCard(DefaultCardInfo):

    #PIPEFY FIELDS
    solicita_o_comprador_pedido            = CardField(str,default="Perda Financeira Plataforma")
    teste_quem_o_respons_vel_pela_perda    = CardField(str,default="Sympla - Cobranças")
    por_que_a_sympla_deve_absorver_a_perda = CardField(str,default="Organizador devendo valor a Sympla")
    como_o_reembolso_deve_acontecer        = CardField(str,default="Transferência bancária ")  #NAO RETIRAR O ESPAÇO
    cpf_ou_cnpj_apenas_n_meros             = CardField(str,default="null")
    copy_of_valor_do_reembolso_obrigat_rio = CardField(float)
    n_do_pedido_obrigat_rio                = CardField(str)


    def __init__(self,**kwargs):
        pass

class Dias180Card(DefaultCardInfo):

    #PIPEFY FIELDS
    solicita_o_comprador_pedido     = CardField(str,default="Reembolso > 180 dias")
    cpf                             = CardField(str,default="null")
    valor_do_reembolso_obrigat_rio  = CardField(float)
    copy_of_n_do_pedido_obrigat_rio = CardField(str)


    #def __init__(self,**kwargs): NewCard.__init__(self,**kwargs)

myNewCard = PerdaFinanceiraCard(
    n_do_evento_obrigat_rio                = "2310947",
    tipo_de_conta                          = "Conta-corrente",
    nome_da_pessoa_titular_da_conta        = "Hannalice Gottschalck Cavalcanti",
    #cnpj                                   = "",
    cpf_ou_cnpj_apenas_n_meros             = "46629599420",
    banco                                  = "Banco do Brasil",
    c_digo_do_banco_n_o_obrigat_rio        = "1",
    ag_ncia_sem_digito_obrigat_rio         = "3525",
    conta_sem_digito_obrigat_rio           = "16276",
    digito_da_conta_obrigat_rio            = "0",
    copy_of_valor_do_reembolso_obrigat_rio = 80,
    n_do_pedido_obrigat_rio                = "26GP37091SR"
)

# myNewCard = Dias180Card(
#     n_do_evento_obrigat_rio         = "2310947",
#     tipo_de_conta                   = "Conta-corrente",
#     nome_da_pessoa_titular_da_conta = "Hannalice Gottschalck Cavalcanti",
#     banco                           = "Banco do Brasil",
#     c_digo_do_banco_n_o_obrigat_rio = "1",
#     ag_ncia_sem_digito_obrigat_rio  = "3525",
#     conta_sem_digito_obrigat_rio    = "16276",
#     digito_da_conta_obrigat_rio     = "0",
#     #cnpj                           = "",
#     cpf                             = "46629599420",
#     valor_do_reembolso_obrigat_rio  = 123.456,
#     copy_of_n_do_pedido_obrigat_rio = "123456"
# )



# cd_id = pfy.createCard(card=myNewCard)


# pfy.deleteCard(cd_id.get("id"))



# pfy.listMembers()

# pfy.listStartFormFields(pipe_id="1025104")

# pfy.listCardsFromPhase(phase_id="325918208")



class MyCard(NewCard):
    #DEFAULT
    __pipeid__               = "302671790"
    __title__                = "Teste"

    #PIPEFY FIELDS
    rfp                                           = CardField(str)
    product_service_needed                        = CardField(str)
    describe_your_request                         = CardField(str)
    define_the_requirements_of_the_ideal_supplier = CardField(str)
    due_date                                      = CardField(str)
    assignee_name                                 = CardField(str)
    primary_contact_email                         = CardField(str)
    copy_of_primary_contact_email                 = CardField(str)
    documents                                     = CardField(list,is_file_path=True)

    


files = [
"D:/Git/Orgs/Dumonde/BidControlSystem/backend/rfps/rfps_baixadas/W912JV-24-Q0031/2024-07-24 1645/Attachment 1 Medford YRRP PWS V2.pdf",
]

myNewCard = MyCard(
      rfp                                           = "TEST",
      product_service_needed                        = "TEST",
      describe_your_request                         = "OLÁ",
      define_the_requirements_of_the_ideal_supplier = "TEST",
      due_date                                      = "2024-01-01 00:00",
      assignee_name                                 = "TEST",
      primary_contact_email                         = "TEST",
      copy_of_primary_contact_email                 = "TEST",
      documents                                     = []
  )

pfy.createCard(card=myNewCard)


pass




from setuptools import setup, find_packages, Extension
 
classifiers = [
  'Development Status :: 5 - Production/Stable',
  'Intended Audience :: Education',
  'Operating System :: Microsoft :: Windows :: Windows 10',
  'License :: OSI Approved :: MIT License',
  'Programming Language :: Python :: 3'
]
 
setup(
  name                          = 'pipefyUts',
  version                       = '1.1.7',
  description                   = 'Pipefy manipulator',
  long_description              = open('README.md').read() + '\n\n' + open('CHANGELOG.txt').read(),
  long_description_content_type = 'text/markdown',
  url                           = '',
  author                        = 'Melque Lima',
  author_email                  = 'melque_ex@yahoo.com.br',
  license                       = 'MIT',
  classifiers                   = classifiers,
  keywords                      = 'pipefyUts',
  packages                      = find_packages(),
  package_data                  = {'': ['*.gql']},
  include_package_data          = True,
  install_requires              = ["requests"]
)
